/*
  t2stars

  Copyright 2009 Jere Leppnen <jere.leppanen@gmail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <libgen.h>

#define LOG(level, args...) do { if(par.verbose>level) { printf("%s(%d): ", __FUNCTION__, __LINE__); printf(args); printf("\n"); } } while(0)
#define ERROR(args...) do { printf("%s(%d): ERROR: ", __FUNCTION__, __LINE__); printf(args); printf("\n"); exit(1); } while(0)

#define PROGRAM "t2stars"

#define PATCHV 1000000.0

#define DEF_FILE "FAM/skyhw/STARS.BIN"
#define DEF_INFILE "tycho2.dat"

uint32_t numw;

/* 
   The dat file contains uint32_t num, then num instances of struct datfmt.
   struct datfmt contains right ascension, declination and magnitude.
*/
uint32_t num=0;
struct datfmt
{
  float a, d, m;
} *dat=NULL;

struct vector
{
  float x, y, z;
} east={0, 1, .9},
  north={1, 0, 0}, 
  up={0, .9, -1};

struct star
{
  struct vector v;
  float b;
};

struct
{
  int verbose, num, lines;
  float lst, month, time, latitude, dip, atm, brightness, multiplier, gamma;
  struct vector lat, rot;
  char *file;
  FILE *infile;
} par={0, -1, 0, 
       0, 5, 21.3233, 0, 120, .25, 0, 1, 1,
       {0, .9, -1}, {1, 0, 0}, 
       NULL, NULL};

void normalize(struct vector *v)
{
  LOG(3, " ");
  float d=sqrt(v->x*v->x+v->y*v->y+v->z*v->z);
  if(d) { v->x/=d; v->y/=d; v->z/=d; }
}

inline float mtob(float m)
{
  LOG(3, " ");
  return pow((pow(1.8, 16-m)*.0004-.04), par.gamma)*par.multiplier+par.brightness;
}

void transform(struct vector *v)
{
  LOG(3, " ");

  float x, y, z;

  y=v->y*par.lat.y-v->z*par.lat.z;
  z=v->y*par.lat.z+v->z*par.lat.y;
  v->y=y;
  v->z=z;

  x=v->x*par.rot.x-v->y*par.rot.y;
  y=v->x*par.rot.y+v->y*par.rot.x;
  v->x=x;
  v->y=y;
}

void makestar(const struct datfmt *d, FILE *file)
{
  LOG(3, " ");
  struct star s={
    {
      cos(M_PI*(d->a+270.0-par.lst)/180.0)*cos(M_PI*d->d/180.0),
      sin(M_PI*(d->a+270.0-par.lst)/180.0)*cos(M_PI*d->d/180.0),
      sin(M_PI*d->d/180.0)
    },
    mtob(d->m)};
  transform(&s.v);
  if(par.dip>0)
    {
      float dimf=(1.0+cos(par.dip/180.0*M_PI))*.5;
      if((1.0+s.v.z)*.5<dimf) dimf=0;
      else dimf=((1.0+s.v.z)*.5-dimf)/(1.0-dimf);
      s.b*=pow(dimf, 1.0+par.atm);
    }
  if(s.b<.01) return;
  if(s.b>1) s.b=1;
  struct star sw={
    {
      s.v.x*east.x+s.v.y*north.x+s.v.z*up.x,
      s.v.x*east.y+s.v.y*north.y+s.v.z*up.y,
      s.v.x*east.z+s.v.y*north.z+s.v.z*up.z
    },
    -s.b};
  if(fwrite(&sw, sizeof(sw), 1, file)!=1) ERROR("Cannot write to file. Disk full?");
  numw++;
}

void readdat()
{
  LOG(2, " ");
  if(!par.infile) ERROR("Cannot open input file.");
  if(fread(&num, sizeof(num), 1, par.infile)!=1) ERROR("Cannot read input file.");
  if(!(dat=(struct datfmt *)malloc(sizeof(struct datfmt)*num))) ERROR("Cannot allocate %d stars.", num);
  if(fread(dat, sizeof(struct datfmt), num, par.infile)!=num) ERROR("Cannot read %d stars from input file.", num);
}

int fileexists(const char *name)
{
  LOG(2, " ");
  FILE *file=fopen(name, "rb");
  if(file) fclose(file);
  LOG(1, "File %s %s.", name, file ? "exists" : "doesn't exist");
  return file!=NULL;
}

void searchdir()
{
  LOG(2, " ");

  int c;
  char *cdir="", *ldir=NULL;
  for(c=0; c<128; c++)
    {
      free(ldir);
      ldir=cdir;
      if(!(cdir=getcwd(NULL, 0))) ERROR("getcwd()");
      LOG(3, "%s", cdir);
      if(!strcmp(ldir, cdir)) break;
      if(fileexists("thief2.exe"))
	{
	  printf("Thief2 directory is %s.\n", cdir);
	  return;
	}
      chdir("..");
    }
  ERROR("Cannot find Thief2 installation directory.");
}

int confirm(const char *msg)
{
  LOG(2, " ");

  printf(msg);
  fflush(stdout);
  char tmp[128];
  memset(tmp, 0, 128);
  scanf("%s", tmp);
  if(*tmp=='y' || *tmp=='Y') return 1;
  return 0;
}

void createdir(const char *file)
{
  LOG(0, "Creating directory for file %s.", file);
  
  char *dir=strdup(file);
  char *p=dir;
  while((p=strpbrk(p, "/\\")))
    {
      *p=0;
      errno=0;
      int r=mkdir(dir);
      LOG(1, "mkdir(%s)=%d (%s)", dir, r, strerror(errno));
      *p='/';
      p++;
    }
}

void cmd_genstars()
{
  LOG(2, " ");

  LOG(2, "east={%g, %g, %g}", east.x, east.y, east.z);
  LOG(2, "north={%g, %g, %g}", north.x, north.y, north.z);
  LOG(2, "up={%g, %g, %g}", up.x, up.y, up.z);
  LOG(2, "lat={%g, %g, %g}", par.lat.x, par.lat.y, par.lat.z);
  LOG(2, "rot={%g, %g, %g}", par.rot.x, par.rot.y, par.rot.z);

  readdat();

  if(!par.file) par.file=DEF_FILE;
  printf("Generating %s:\n", par.file);
  if(fileexists(par.file))
    {
      if(!confirm("File exists and will be overwritten. Are you sure you want to permanently\n"
		  "destroy the existing file and generate a new one (y/n)? "))
	{
	  printf("No changes made to %s.\n", par.file);
	  par.file=NULL;
	  return;
	}
    }
  createdir(par.file);
  FILE *file=fopen(par.file, "wb");
  if(!file) ERROR("Cannot open %s for writing.", par.file);
  numw=0;
  if(fwrite(&numw, sizeof(numw), 1, file)!=1) ERROR("Cannot write to %s.", par.file);

  struct datfmt ds;

  if(par.lines)
    {
      LOG(0, "Generating lines.");
      float a, d;
      ds.m=0;
      for(d=-80; d<=80; d+=10.0)
	{
	  for(a=0; a<360; a+= (d==0 ? .1 : 1.0) )
	    {
	      ds.a=a;
	      ds.d=d;
	      makestar(&ds, file);
	    }
	}
      for(a=0; a<360; a+=10.0)
	{
	  for(d=-89; d<=89; d+= (a==0 ? .2 : 1.0) )
	    {
	      ds.a=a;
	      ds.d=d;
	      makestar(&ds, file);
	    }
	}
      for(a=0; a<5; a+=.1)
	{
	  ds.a=a;
	  ds.d=65;
	  makestar(&ds, file);
	}
      for(d=60; d<65; d+=.1)
	{
	  ds.a=15;
	  ds.d=d;
	  makestar(&ds, file);
	}
      ds.a=0;
      ds.d=90;
      makestar(&ds, file);
      ds.d=-90;
      makestar(&ds, file);
    }

  LOG(0, "Generating stars.");
  int c;
  for(c=0; c<num && (par.num==-1 || numw<par.num); c++)
    makestar(dat+c, file);

  if(fseek(file, 0, SEEK_SET)) ERROR("Cannot seek to beginning of %s.", par.file);
  if(fwrite(&numw, sizeof(numw), 1, file)!=1) ERROR("Cannot write to %s.", par.file);
  fclose(file);
  printf("%s created, %d stars.\n", par.file, numw);
  par.file=NULL;
}

void cmd_help()
{
  LOG(2, " ");
  printf("\nUsage: %s [command] [options]\n"
	 "\nCommands:\n\n"
	 "    genstars  Generate STARS.BIN.\n\n"
	 "    misinfo   Show the current \"max sky intensity\" value in *.mis file(s).\n"
         "              Nothing to do with misinformation.\n\n"
	 "    mispatch  Patch the \"max sky intensity\" value in *.mis file(s).\n\n"
	 "\nOptions:\n"
	 "    -a <float>  latitude (%g)\n"
	 "    -b <float>  brightness (%g)\n"
	 "    -d <float>  horizon dip angle (%g)\n"
	 "    -f <file>   operate on specified file\n"
	 "    -g <float>  gamma (%g)\n"
	 "    -h          help\n"
	 "    -i <file>   input file (\"%s\")\n"
	 "    -l          longitudinal and latitudinal lines\n"
	 "    -m <float>  month of the year (%g)\n"
	 "    -n <int>    number of stars\n"
	 "    -o <float>  atmosphere density (%g)\n"
	 "    -r <float>  clockwise rotation (0)\n"
	 "    -s <float>  local sidereal time in degrees (%g)\n"
	 "    -t <float>  time of day, hours (%g)\n"
	 "    -u <float>  multiplier (%g)\n"
	 "    -v <int>    verbosity level (%d)\n",
	 PROGRAM, par.latitude, par.brightness, par.dip, par.gamma, DEF_INFILE,
	 par.month, par.atm, par.lst, par.time, par.multiplier, par.verbose);
  printf("\n\
This program replaces the stock stars in Thief 2: The Metal Age with\n\
realistic stars. (For most missions; in some of the OMs, the compass\n\
doesn't point north.)\n\
\n\
The easiest way to use this program is to place it anywhere under your\n\
Thief2 installation directory and then just double-click on\n\
t2stars.exe. You'll be prompted on each file that is patched or\n\
replaced.\n\
\n\
Only a single command may be specified. If no command is given, the\n\
default is to execute both genstars and mispatch. Options may appear\n\
in any order, the last of conflicting options takes effect.\n\
\n\
Thief2 installation directory is searched by starting at the current\n\
working directory and going up the directory tree level by level until\n\
thief2.exe is found. Any relative paths used with -f and -i options\n\
are relative to this directory.\n\
\n\
Local time is specified with the -m and -t options. For example, -m 1\n\
is Jan 1st. -m 1.5 is Jan 16th. -t 21.5 is half past nine in the\n\
evening, in mean solar time. Alternatively, local sidereal time can be\n\
specified directly with the -s option.\n\
\n\
Star brightness is controlled by the brightness, multiplier and gamma\n\
options. The resulting brightness, given a base brightness in the range\n\
0-1, is (base^gamma)*multiplier+brightness.\n\
\n\
Example: Generate an almost exact match with the original stars:\n\
\n\
    t2stars genstars -n 1500 -r 90\n\
\n\
Example: The starry sky where and when the initial version of this\n\
program was written (Helsinki, Dec 23rd, 22:15):\n\
\n\
    t2stars genstars -a 60.17 -m 12.73 -t 22.25\n\
\n");
  exit(1);
}

#define BUFSIZE 4096

int scanmis(const char *name, size_t *offset, float *f)
{
  LOG(2, "name=%s", name);

  *offset=0;
  *f=0;
  
  FILE *file=fopen(name, "rb");
  if(!file) ERROR("Cannot open file %s.", name);
  
  char *buf=malloc(BUFSIZE*2);
  memset(buf, 0, BUFSIZE*2);

  const char *needle="STAROBJVAR";

  int block=-1;
  while(!feof(file))
    {
      memcpy(buf, buf+BUFSIZE, BUFSIZE);
      int i=fread(buf+BUFSIZE, 1, BUFSIZE, file);
      memset(buf+BUFSIZE+i, 0, BUFSIZE-i);
      char *p=memchr(buf, *needle, BUFSIZE);
      while(p)
	{
	  if(!memcmp(needle, p, strlen(needle)))
	    {
	      *offset=block*BUFSIZE+(p-buf);
	      LOG(1, "Found %s at 0x%x", needle, *offset);
	      p+=40;
	      *f=*((float *)p);
	      return 1;
	    }
	  p=memchr(p+1, *needle, BUFSIZE-(p-buf)-1);
	}
      block++;
    }
  fclose(file);
  return 0;
}

int misinfo(const char *name)
{
  LOG(2, " ");

  size_t offset;
  float f;
  if(scanmis(name, &offset, &f))
    {
      printf("%s NewSky:Stars / Max Sky Intensity: %g %s\n", name, f, f==PATCHV ? "(patched value)" : "");
      return 0;
    }
  
  LOG(0, "%s doesn't look like a mis file.", name);
  return 1;
}

void cmd_misinfo()
{
  LOG(2, " ");

  if(par.file)
    {
      if(misinfo(par.file)) ERROR("%s doesn't look like a mis file.", par.file);
      return;
    }

  DIR *dir=opendir(".");
  if(!dir) ERROR("Cannot open current directory.");
  struct dirent *ent;
  while((ent=readdir(dir)))
    {
      LOG(3, "dirent=%s", ent->d_name);
      if(!strcasecmp(ent->d_name+strlen(ent->d_name)-4, ".mis"))
	misinfo(ent->d_name);
    }
}

void mispatch(const char *name)
{
  LOG(2, " ");

  size_t offset;
  float f, fm=PATCHV;
  if(scanmis(name, &offset, &f))
    {
      if(f==fm) LOG(0, "%s already patched - no action needed", name);
      else
	{
	  printf("\nPatching file %s:\n", name);
	  if(confirm("Patching is not undoable - information will be lost (in most cases unused\n"
		     "information, but still ... you've been warned).\n\n"
		     "Are you sure you've made backups and want to continue (y/n)? "))
	    {
	      FILE *file=fopen(name, "r+b");
	      if(!file) ERROR("Cannot open %s for writing.", name);
	      if(fseek(file, offset+40, SEEK_SET)) ERROR("Cannot seek to 0x%x in %s.", offset+40, name);
	      if(fwrite(&fm, sizeof(float), 1, file)!=1) ERROR("Cannot write to %s.", name);
	      printf("%s patched.\n", name);
	      return;
	    }
	  printf("No changes made to %s.\n", name);
	}
    }
  else 
    LOG(0, "%s doesn't look like a mis file.", name);
}

void cmd_mispatch()
{
  LOG(2, " ");

  if(par.file)
    {
      mispatch(par.file);
      return;
    }

  DIR *dir=opendir(".");
  if(!dir) ERROR("Cannot open current directory.");
  struct dirent *ent;
  while((ent=readdir(dir)))
    {
      LOG(3, "dirent=%s", ent->d_name);
      if(!strcasecmp(ent->d_name+strlen(ent->d_name)-4, ".mis"))
	mispatch(ent->d_name);
    }
}

void chkw()
{
  LOG(2, " ");

  if(sizeof(float)!=4)
    ERROR("sizeof(float)=%d != 4", sizeof(float));
  if(sizeof(struct star)!=16)
    ERROR("sizeof(struct star)=%d != 16", sizeof(struct star));
}

void setlst()
{
  LOG(2, " ");
  int y=2000, m=par.month, day=1+(int)((par.month-(float)m)*30.0);
  par.lst=280.461+360.986*
    ((float)(367*y-(7*(y+((m+9)/12))/4)+(275*m/9)+day)-730531.5+par.time/24.0);
  par.lst=fmodf(par.lst, 360.0);
  LOG(1, "lst=%g", par.lst);
}

int main(int argc, char *argv[])
{
  char *inname=NULL;
  float r;

  normalize(&east);
  normalize(&north);
  normalize(&up);
  normalize(&par.lat);
  normalize(&par.rot);

  setlst();
  par.latitude=acos(east.y)/M_PI*180.0;

  int c;
  for(c=1; c<argc; c++)
    {
      if(!strncasecmp(argv[c], "-h", 2)) cmd_help();
    }

  opterr=0;
  while(1)
    {
      int o=getopt(argc, argv, "a:b:d:f:g:i:lm:n:o:r:s:t:u:v:");
      if(o<0) break;
      switch(o)
	{
	case 'a':
	  par.latitude=(90.0-atof(optarg))/180.0*M_PI;
	  par.lat.y=cos(par.latitude);
	  par.lat.z=-sin(par.latitude);
	  break;
	case 'b':
	  par.brightness=atof(optarg);
	  break;
	case 'd':
	  par.dip=atof(optarg);
	  break;
	case 'f':
	  par.file=strdup(optarg);
	  break;
	case 'g':
	  par.gamma=atof(optarg);
	  break;
	case 'i':
	  inname=strdup(optarg);
	  break;
	case 'l':
	  par.lines=1;
	  break;
	case 'm':
	  par.month=atof(optarg);
	  setlst();
	  break;
	case 'n':
	  par.num=atoi(optarg);
	  break;
	case 'o':
	  par.atm=atof(optarg);
	  break;
	case 'r':
	  r=atof(optarg)/180.0*M_PI;
	  par.rot.x=cos(r);
	  par.rot.y=-sin(r);
	  break;
	case 's':
	  par.lst=atof(optarg);
	  break;
	case 't':
	  par.time=atof(optarg);
	  setlst();
	  break;
	case 'u':
	  par.multiplier=atof(optarg);
	  break;
	case 'v':
	  par.verbose=atoi(optarg);
	  break;
	case ':': case '?': default:
	  ERROR("Invalid option \'%c\' or missing value. (Try %s -h)", optopt, PROGRAM);
	}
    }

  LOG(1, "%s %s", __FILE__, __DATE__);

  if(!inname)
    {
      inname=malloc(strlen(argv[0])+strlen(DEF_INFILE));
      strcpy(inname, argv[0]);
      char *p=inname+strlen(argv[0]);
      while(p>inname && *p!='/' && *p!='\\') p--;
      if(*p=='/' || *p=='\\') p++;
      strcpy(p, DEF_INFILE);
    }
  
  LOG(1, "Input file is %s.", inname);
  par.infile=fopen(inname, "rb");

  chkw();

  searchdir();

  if(optind<argc)
    {
      if(!strcasecmp(argv[optind], "genstars")) cmd_genstars();
      else if(!strcasecmp(argv[optind], "misinfo")) cmd_misinfo();
      else if(!strcasecmp(argv[optind], "mispatch")) cmd_mispatch();
      else ERROR("Invalid command %s. (Try %s -h)", argv[optind], PROGRAM);
      return 0;
    }

  cmd_genstars();
  cmd_mispatch();

  return 0;
}
